import ray
from ray.util.queue import Queue

@ray.remote
class QueuedSendReceive(object):
    """
    Send and receive through queues.
    Queues make the send call asynchronous.

    # RB note the queue breaks the send/send deadlock
    # RB note how many queues are needed?  why inferior to push model
    """
    msg = ""

    def __init__(self, squeue, rqueue):
        """Create an inbound and outbound queue for each actor."""
        self.recvQ = rqueue
        self.sendQ = squeue

    def set_message(self, msg):
        """set the message to be sent"""
        self.msg = msg

    def _qsend(self):
        """Helper: enqueue a message"""
        ### implement if you want or delete function

    def _qreceive(self):
        """Helper: dequeue a message"""
        ### implement if you want or delete function

    def rs_exchange(self):
        """receive first then send"""
        ### TODO

    def sr_exchange(self):
        """send first then receive"""
        ### TODO

# script to drive parallel program
ray.init(num_cpus=4, ignore_reinit_error=True)

# create messaging queues
q0to1 = Queue(maxsize=100)
q1to0 = Queue(maxsize=100)

# objects with paired queues
sr0 = QueuedSendReceive.remote(q0to1,q1to0)
sr1 = QueuedSendReceive.remote(q1to0,q0to1)

# set messages
roid0 = sr0.set_message.remote(f"Message from {sr0}")
roid1 = sr1.set_message.remote(f"Message from {sr1}")
ray.get(roid0)
ray.get(roid1)

print("Send first. No deadlock")
oid0 = sr0.sr_exchange.remote()
oid1 = sr1.sr_exchange.remote()
print(ray.get(oid0))
print(ray.get(oid1))

print("Paired. One sender, one receiver. No deadlock.")
oid0 = sr0.sr_exchange.remote()
oid1 = sr1.rs_exchange.remote()
print(ray.get(oid0))
print(ray.get(oid1))

print("Receive first, deadlock")
oid0 = sr0.rs_exchange.remote()
oid1 = sr1.rs_exchange.remote()
print(ray.get(oid0))
print(ray.get(oid1))

print("Didn't get here. Deadlock.")

ray.shutdown()
